import numpy as np
import pyqg
import pytest
import os
import pickle
from pyqg.diagnostic_tools import calc_ispec

def test_calc_ispec_peak():
    # Create a radial sine wave spiraling out from the center of the model's
    # spatial field (with a given frequency)
    m = pyqg.QGModel()
    radius = np.sqrt((m.x-m.x.mean())**2 + (m.y-m.y.mean())**2)
    frequency = m.k[0][20]
    radial_sine = np.sin(radius * frequency)

    # Take its FFT
    radial_sine_fft = m.fft(np.array([radial_sine, radial_sine]))[0]

    # Compute an isotropic spectrum
    iso_wavenumbers, iso_spectrum = calc_ispec(m, radial_sine_fft)

    # Its peak should be at the closest frequency to the true frequency
    spectrum_peak_idx = np.argmax(iso_spectrum)
    sinewave_freq_idx = np.argmin(np.abs(iso_wavenumbers - frequency))
    assert spectrum_peak_idx == sinewave_freq_idx

def test_calc_ispec_units(rtol=1e-5):
    fixtures_path = f"{os.path.dirname(os.path.realpath(__file__))}/fixtures"

    with open(f"{fixtures_path}/LayeredModel_params.pkl", 'rb') as f:
        # Common set of parameters for each model
        params = pickle.load(f)

    m1 = pyqg.LayeredModel(nx=96, **params)
    m2 = pyqg.LayeredModel(nx=64, **params)
    m1.q = np.load(f"{fixtures_path}/LayeredModel_nx96_q.npy")
    m2.q = np.load(f"{fixtures_path}/LayeredModel_nx64_q.npy")
    for m in [m1, m2]:
        m._invert()
        m._calc_derived_fields()
        for a in ['q','p']:
            for z in [0,1]:
                signal2d = getattr(m, a)[z]
                spectral = getattr(m, a+'h')[z]
                power = np.abs(spectral)**2/m.M**2
                k, ispec = calc_ispec(m, power, averaging=False)
                dk = k[1]-k[0]
                np.testing.assert_allclose(
                    signal2d.var(),
                    ispec.sum()*dk,
                    rtol,
                    err_msg=f"ispec should have correct integral for {a}{z+1}"
                )

def test_calc_ispec_sum(): 

    for nx in [16, 64, 256]:
        ny = nx
        m = pyqg.QGModel(nx = nx, ny = ny) 
        p = np.random.rand(ny, nx)

        # Get energy field and its spectrum
        E = np.abs(p)**2 # Energy in real space
        Eh_numpy = np.abs(np.fft.fft2(p))**2 # Energy spectrum in full plane
        Eh_model = np.abs(m.fft(p.reshape(1, nx, ny))[0])**2 # Energy spectrum in model (half) plane

        # Do the same in calc_ispec to avoid double counting 0th wavenumber and the largest wavenumber
        Eh_model_2 = Eh_model.copy()
        Eh_model_2[...,0] /= 2
        Eh_model_2[...,-1] /= 2

        # Get variance (or average energy)
        E_total = E.mean()

        # Check that the half and full planes both satisfy Parseval's theorem
        np.testing.assert_allclose(E_total, Eh_numpy.sum()/m.M**2)
        np.testing.assert_allclose(E_total, Eh_model_2.sum()/m.M**2*2) # Note the factor of 2 here

        ## Test calc_ispec()
        # Check that Parseval's theorem is roughly satisfied with averaging and truncation
        kr, Ehr = calc_ispec(m, Eh_model, truncate=True, averaging=True)
        E_total_radial = np.cumsum(Ehr * (kr[1]-kr[0]))[-1]
        assert E_total_radial/Eh_numpy.sum() > 0.5 and E_total_radial/Eh_numpy.sum() < 2,\
            f"Parseval's theorem is not roughly satisfied by calc_ispec"

        # Check that Parseval's theorem is exactly satisfied without averaging or truncation
        kr, Ehr = calc_ispec(m, Eh_model, truncate=False, averaging=False)
        E_total_radial = Ehr.sum() * (kr[1]-kr[0])
        np.testing.assert_allclose(E_total_radial, Eh_numpy.sum())

if __name__ == "__main__":
    test_calc_ispec_peak()
    test_calc_ispec_units()
    test_calc_ispec_sum()



import numpy as np
import pyqg
import xarray as xr
import pytest

year = 24*60*60*360.

expected_vars = [
    'q', 
    'u', 
    'v', 
    'ufull', 
    'vfull', 
    'p',
    'dqdt',
    'qh', 
    'uh', 
    'vh', 
    'ph', 
    'Ubg', 
    'Qy',
    'dqhdt',
]

expected_diags = [
    'EKE', 
    'entspec', 
    'APEflux', 
    'APEflux_div', 
    'KEflux', 
    'KEflux_div', 
    'APEgen',
    'APEgenspec',
    'Ensspec',
    'KEspec',
    'EKEdiss',
    'KEspec_modal',
    'PEspec_modal',
    'APEspec',
    'ENSflux',
    'ENSgenspec',
    'paramspec',
    'paramspec_APEflux',
    'paramspec_KEflux',
    'Dissspec',
    'KEfrictionspec',
    'ENSDissspec',
    'ENSfrictionspec',
]

expected_attrs = [
    'L',
    'W',
    'dt',
    'filterfac',
    'nk',
    'nl',
    'ntd',
    'nx',
    'ny',
    'nz',
    'rek',
    'taveint',
    'tavestart',
    'tc',
    'tmax',
    'twrite',
]

expected_coords = [
    'time',
    'lev',
    'x',
    'y',
    'l',
    'k',
]

def QG():
    '''Initialize Two-layer Model'''
    return pyqg.QGModel(tmax=year/2, twrite=1000, tavestart=year/3)

def Layered():
    '''Initialize Layered quasigeostrophic model'''
    L =  1000.e3     # length scale of box    [m]
    Ld = 15.e3       # deformation scale      [m]
    kd = 1./Ld       # deformation wavenumber [m^-1]
    Nx = 64          # number of grid points

    H1 = 500.        # layer 1 thickness  [m]
    H2 = 1750.       # layer 2
    H3 = 1750.       # layer 3

    U1 = 0.05          # layer 1 zonal velocity [m/s]
    U2 = 0.025         # layer 2
    U3 = 0.00          # layer 3

    rho1 = 1025.
    rho2 = 1025.275
    rho3 = 1025.640

    rek = 1.e-7       # linear bottom drag coeff.  [s^-1]
    f0  = 0.0001236812857687059 # coriolis param [s^-1]
    beta = 1.2130692965249345e-11 # planetary vorticity gradient [m^-1 s^-1]

    return pyqg.LayeredModel(nx=Nx, nz=3, U=[U1,U2,U3], V=[0.,0.,0.], L=L, f=f0, beta=beta,
                             H=[H1,H2,H3], rho=[rho1,rho2,rho3], rek=rek,
                             tmax=year/2, twrite=1000, tavestart=year/3)

def BT():
    '''Initialize Barotropic model'''
    return pyqg.BTModel(L=2.*np.pi, nx=256, beta=0., H=1., rek=0., 
                        rd=None, tmax=year/2, twrite=1000, tavestart=year/3,
                        dt=year/8, taveint=1, ntd=4)
    
def SQG():
    '''Initialize Surface Quasi-Geostrophic Model'''
    return pyqg.SQGModel(L=2.*np.pi, nx=512, tmax=year/2, beta=0.,
                         Nb=1., H=1., f_0 = 1., dt=year/8,
                         taveint=1, twrite=1000, ntd=4, tavestart=year/3)
    
@pytest.fixture(params=[QG, Layered, BT, SQG])
def all_models(request):
    model = request.param()
    return model

def test_xarray(all_models):
    '''Run with snapshots and test contents of xarray.dataset'''
    m=all_models
    tsnapint=year/4
    datasets = []
    timevals = []

    for snapshot in m.run_with_snapshots(tsnapstart=m.t, tsnapint=tsnapint):
        ds = m.to_dataset()
        assert type(ds) == xr.Dataset

        datasets.append(ds)
        timevals.append(m.t)

        if snapshot < tsnapint:

            for v in expected_vars:
                assert v in ds

            for a in expected_attrs:
                assert f"pyqg:{a}" in ds.attrs

            for c in expected_coords:
                assert c in ds.coords

        if snapshot > tsnapint:

            for v in list(ds.keys()): 
                assert v in expected_vars + expected_diags

            for v in expected_diags:
                if v in ds:
                    assert 'time' in ds[v].coords

            for a in expected_attrs:
                assert f"pyqg:{a}" in ds.attrs

            for c in expected_coords:
                assert c in ds.coords

    concatenated = xr.concat(datasets, dim='time')

    assert np.allclose(concatenated.coords['time'], timevals)



import numpy as np
from numpy import pi
from . import model

class QGDiagnostics(model.Model):
    def _initialize_model_diagnostics(self):
        self.add_diagnostic('entspec',
                description='barotropic enstrophy spectrum',
                function= (lambda self:
                    np.abs((self.Hi[:,np.newaxis,np.newaxis]*self.qh).sum(axis=0)/self.H)**2/self.M**2),
                units='m s^-2',
                dims=('l','k')
        )

        self.add_diagnostic('paramspec_APEflux',
            description='total additional APE flux due to subgrid parameterization',
            function=(lambda self: 
                -self._calc_paramspec_contribution(np.einsum("ij, jk..., k... -> i...", 
                        self.S, self.a, self._calc_parameterization_contribution()))
                ),
            units='m^2 s^-3',
            dims=('l','k')
        )

        self.add_diagnostic('paramspec_KEflux',
            description='total additional KE flux due to subgrid parameterization',
            function=(lambda self: 
                self.wv2*self._calc_paramspec_contribution(np.einsum("ij..., j... -> i...", 
                        self.a, self._calc_parameterization_contribution()))
                ),
            units='m^2 s^-3',
            dims=('l','k')
        )

        self.add_diagnostic('ENSflux',
                 description='barotropic enstrophy flux',
                 function = (lambda self: (-self.Hi[:,np.newaxis,np.newaxis]*
                              (self.qh.conj()*self.Jq).real).sum(axis=0)/self.H/self.M**2),
                units='s^-3',
                dims=('l','k')
        )

        self.add_diagnostic('ENSgenspec',
                    description='the spectrum of the rate of generation of barotropic enstrophy',
                    function = (lambda self:
                                (self.Hi[:,np.newaxis,np.newaxis]*((self.ilQx-self.ikQy)*
                                self.Sph.conj()*self.ph).real).sum(axis=0)/self.H/self.M**2),
                units='s^-3',
                dims=('l','k')
        )

        self.add_diagnostic('ENSfrictionspec',
                    description='the spectrum of the rate of dissipation of barotropic enstrophy due to bottom friction',
                    function = (lambda self: self.rek*self.Hi[-1]/self.H*self.wv2*
                        (self.qh[-1].conj()*self.ph[-1]).real/self.M**2), 
                units='s^-3',
                dims=('l','k')
        )

        self.add_diagnostic('APEgenspec',
                    description='the spectrum of the rate of generation of available potential energy',
                    function =(lambda self: (self.Hi[:,np.newaxis,np.newaxis]*
                                (self.Ubg[:,np.newaxis,np.newaxis]*self.k +
                                 self.Vbg[:,np.newaxis,np.newaxis]*self.l)*
                                (1j*self.ph.conj()*self.Sph).real).sum(axis=0)/self.H/self.M**2),
                units='m^2 s^-3',
                dims=('l','k')
        )

    def _calc_paramspec_contribution(self, term):
        height_ratios = (self.Hi/self.H)[:,np.newaxis,np.newaxis]
        return np.real(height_ratios*self.ph.conj()*term).sum(axis=0) / self.M**2

    def _calc_cfl(self):
        return np.abs(
            np.hstack([self.u + self.Ubg[:,np.newaxis,np.newaxis], self.v])
        ).max()*self.dt/self.dx

    # calculate KE: this has units of m^2 s^{-2}
    #   (should also multiply by H1 and H2...)
    def _calc_ke(self):
        ke = 0.
        for j in range(self.nz):
            ke += .5*self.Hi[j]*self.spec_var(self.wv*self.ph[j])
        return ke.sum() / self.H

    # calculate eddy turn over time
    # (perhaps should change to fraction of year...)
    def _calc_eddy_time(self):
        """ estimate the eddy turn-over time in days """
        ens = 0.
        for j in range(self.nz):
            ens = .5*self.Hi[j] * self.spec_var(self.wv2*self.ph[j])

        return 2.*pi*np.sqrt( self.H / ens.sum() ) / 86400



import matplotlib.pyplot as plt
import numpy as np
from numpy import pi
import pyqg

# the model object
year = 1.
m = pyqg.BTModel(L=2.*pi,nx=128, tmax = 1*year,
        beta = 20., H = 1., rek = 0., rd = None, dt = 0.001,
                     taveint=year, ntd=4)

# Gaussian IC
fk = m.wv != 0
ckappa = np.zeros_like(m.wv2)
ckappa[fk] = np.sqrt( m.wv2[fk]*(1. + (m.wv2[fk]/36.)**2) )**-1

nhx,nhy = m.wv2.shape

R = pi/6.
Pi = -np.exp(-((m.x-3*pi/2)**2 + (m.y-pi)**2)/R**2)

Pi = Pi - Pi.mean()
Pi_hat = m.fft( Pi[np.newaxis,:,:] )
KEaux = m.spec_var(m.wv*Pi_hat )

pih = ( Pi_hat/np.sqrt(KEaux) )
qih = -m.wv2*pih
qi = m.ifft(qih)
m.set_q(qi)

# run the model
plt.rcParams['image.cmap'] = 'RdBu'

plt.ion()

for snapshot in m.run_with_snapshots(tsnapstart=0, tsnapint=10*m.dt):
    plt.clf()
    plt.imshow(m.q.squeeze())
    plt.clim([-20., 20.])
    plt.xticks([])
    plt.yticks([])

    plt.pause(0.01)
    plt.draw()
    plt.ioff()
    




import unittest
import numpy as np
import pyqg

class LayeredModelTester(unittest.TestCase):

    def setUp(self):

        self.m = pyqg.LayeredModel(
                    nz = 3,
                    U  = [.1,.05,.0],
                    V  = [.1,.05,.0],
                    rho= [.1,.3,.5],
                    H  = [.1,.1,.3],
                    f  = 1.,
                    beta = 0.)

        self.atol=1.e-16

        # creates stretching matrix from scratch
        self.S = np.zeros((self.m.nz,self.m.nz))

        F11 = self.m.f2/self.m.gpi[0]/self.m.Hi[0]
        F12 = self.m.f2/self.m.gpi[0]/self.m.Hi[1]
        F22 = self.m.f2/self.m.gpi[1]/self.m.Hi[1]
        F23 = self.m.f2/self.m.gpi[1]/self.m.Hi[2]

        self.S[0,0], self.S[0,1] = -F11, F11
        self.S[1,0], self.S[1,1], self.S[1,2] = F12, -(F12+F22), F22
        self.S[2,1], self.S[2,2] = F23, -F23

    def test_stretching(self):
        """ Check if stretching matrix is consistent
                and satisfies basic properties """

        # the columns of the S must add to zero (i.e, S is singular)
        err_msg = ' Zero is not an eigenvalue of S'
        assert np.all(self.m.S.sum(axis=1)==0.) , err_msg

        # the matrix Hi * S must by a symmetric matrix
        HS = np.dot(np.diag(self.m.Hi),self.m.S)
        np.testing.assert_allclose(HS,HS.T,atol=self.atol,
                err_msg=' Hi*S is not symmetric')

        np.testing.assert_allclose(self.m.S,self.S,atol=self.atol,
                err_msg= ' Unmatched stretching matrix')

    def test_init_background(self):
        """ Check the initialization of the mean PV gradiends  """

        Qy = -np.einsum('ij,j->i',self.S,self.m.Ubg)
        Qx = np.einsum('ij,j->i',self.S,self.m.Vbg)

        np.testing.assert_allclose(Qy,self.m.Qy,atol=self.atol,
                err_msg=' Unmatched Qy')
        np.testing.assert_allclose(Qx,self.m.Qx,atol=self.atol,
                err_msg=' Unmatched Qx ')


    def test_inversion_matrix(self):
        """ Check the inversion matrix """

        # it suffices to test for a single wavenumber
        M = self.S - np.eye(self.m.nz)*self.m.wv2[5,5]
        Minv = np.zeros_like(M)

        detM = np.linalg.det(M)

        Minv[0,0] = M[1,1]*M[2,2] - M[1,2]*M[2,1]
        Minv[0,1] = M[0,2]*M[2,1] - M[0,1]*M[2,2]
        Minv[0,2] = M[0,1]*M[1,2] - M[0,2]*M[1,1]

        Minv[1,0] = M[1,2]*M[2,0] - M[1,0]*M[2,2]
        Minv[1,1] = M[0,0]*M[2,2] - M[0,2]*M[2,0]
        Minv[1,2] = M[0,2]*M[1,0] - M[0,0]*M[1,2]

        Minv[2,0] = M[1,0]*M[2,1] - M[1,1]*M[2,0]
        Minv[2,1] = M[0,1]*M[2,0] - M[0,0]*M[2,1]
        Minv[2,2] = M[0,0]*M[1,1] - M[0,1]*M[1,0]

        Minv = Minv/detM

        np.testing.assert_allclose(self.m.a[:,:,5,5], Minv,atol=self.atol,
                err_msg= ' Unmatched inversion matrix ')

if __name__ == "__main__":
    unittest.main()



import pyqg
import time
import cProfile
import pstats
import numpy as np

tmax = 8000*1000

dtfac = 64 * 8000.

mynx = [32, 64, 128, 256, 512, 1024, 2048]
mynth = [1,2,4,8,16,32]
res = np.zeros((len(mynx), 5))


print 'nx, threads, timesteps, time'
for j, nx in enumerate(mynx):

    dt = dtfac / nx

    #for i, (use_fftw, nth) in enumerate([(False, 1), (True, 1),
    #    (True, 2), (True, 4), (True, 8)]):
    for i, nth in enumerate(mynth):

        m = pyqg.QGModel(nx=nx, tmax=tmax, dt=dt, ntd=nth,
                         # no output    
                         twrite=np.inf,
                         # no time average
                         taveint=np.inf,)


    
        tic = time.time()
        m.run()
        toc = time.time()
        tottime = toc-tic
        #res[j,i] = tottime
        #print 'nx=%3d, fftw=%g, threads=%g: %g' % (nx, use_fftw, nth, tottime)
        print '%3d, %3d, %8d, %10.4f' % (nx, nth, m.tc, tottime)
    


# # profiling
# prof = cProfile.Profile()
# prof.run('m.run()')
# p = pstats.Stats(prof)
# p.sort_stats('cum').print_stats(0.3)



from future.utils import iteritems
import pytest
import unittest
import numpy as np
import pyqg
import pickle
import os
from pyqg import diagnostic_tools as diag

def test_describe_diagnostics():
    """ Test whether describe_diagnostics runs without error """

    m = pyqg.QGModel(1)
    m.describe_diagnostics()

def old_qgmodel_calc_paramspec(self, dqh1, dqh2):
    del1 = self.del1
    del2 = self.del2
    F1 = self.F1
    F2 = self.F2
    wv2 = self.wv2
    ph = self.ph
    return np.real(
        (del1 / (wv2 + F1 + F2) * (-(wv2 + F2) * dqh1 - F1 * dqh2) * np.conj(ph[0])) +
        (del2 / (wv2 + F1 + F2) * (-F2 * dqh1 - (wv2 + F1) * dqh2) * np.conj(ph[1])) +
        (del1 * F1 / (wv2 + F1 + F2) * (dqh2 - dqh1) * np.conj(ph[0] - ph[1]))
    )

def test_paramspec_decomposition(rtol=1e-10):
    # Initialize a model with a parameterization, step it forward and compute paramspec
    dq = np.random.normal(size=(2,64,64))
    m = pyqg.QGModel(q_parameterization = lambda m: dq)
    m._step_forward()
    m._increment_diagnostics()

    # Compute the parameterization spectrum at least two ways
    height_ratios = (m.Hi / m.H)[:,np.newaxis,np.newaxis]
    dqh = m.fft(dq)
    ps1 = -np.real((height_ratios * np.conj(m.ph) * dqh).sum(axis=0)) / m.M**2
    ps2 = old_qgmodel_calc_paramspec(m, dqh[0], dqh[1]) / m.M**2
    ps3 = m.get_diagnostic('paramspec') 

    # Ensure they're identical
    np.testing.assert_allclose(ps1, ps2, rtol=rtol)
    np.testing.assert_allclose(ps1, ps3, rtol=rtol)

    # Now test it can be decomposed into separate KE and APE components
    apeflux_term = np.einsum("ij, jk..., k... -> i...", m.S, m.a, dqh)
    keflux_term  = np.einsum("ij..., j... -> i...", m.a, dqh)
    height_ratios = (m.Hi/m.H)[:,np.newaxis,np.newaxis]
    paramspec_apeflux = -np.real(height_ratios*m.ph.conj()*apeflux_term).sum(axis=0) / m.M**2
    paramspec_keflux  = m.wv2*np.real(height_ratios*m.ph.conj()* keflux_term).sum(axis=0) / m.M**2
    
    ps4 = paramspec_apeflux + paramspec_keflux
    np.testing.assert_allclose(ps1, ps4, rtol=rtol)

    # Test these terms match the subterms from QGModel
    np.testing.assert_allclose(paramspec_apeflux,
            m.get_diagnostic('paramspec_APEflux'), rtol=rtol)
    np.testing.assert_allclose(paramspec_keflux,
            m.get_diagnostic('paramspec_KEflux'), rtol=rtol)

def test_paramspec_additivity(rtol=1e-10):
    # Test over multiple model classes
    for model_class in [pyqg.QGModel, pyqg.LayeredModel]:
        # Initialize four models with different (deterministic) parameterizations
        m1 = model_class()

        dq = np.random.normal(size=m1.q.shape)
        du = np.random.normal(size=m1.u.shape)
        dv = np.random.normal(size=m1.v.shape)

        m2 = model_class(q_parameterization=lambda m: dq)
        m3 = model_class(uv_parameterization=lambda m: (du,dv))
        m4 = model_class(q_parameterization=lambda m: dq,
                         uv_parameterization=lambda m: (du,dv))

        # Give them the same initial conditions
        for m in [m1,m2,m3,m4]:
            m.q = m1.q

        # Step them forward and manually increment diagnostics
        for m in [m1,m2,m3,m4]:
            m._step_forward()
            m._increment_diagnostics()

        # Unparameterized model should have 0 for its parameterization spectrum
        np.testing.assert_allclose(m1.get_diagnostic('paramspec'), 0., rtol=rtol)

        # Parameterized models should have nonzero values
        for m in [m2,m3,m4]:
            with pytest.raises(AssertionError):
                np.testing.assert_allclose(m.get_diagnostic('paramspec'), 0., rtol=rtol)

        # Model with both parameterizations should have the sum
        np.testing.assert_allclose(
                (m2.get_diagnostic('paramspec') + m3.get_diagnostic('paramspec')),
                m4.get_diagnostic('paramspec'),
                rtol=rtol)

def test_Dissspec_diagnostics(atol=1e-20):

    # Run model for some timesteps
    dt = 3600
    tmax = dt * 1000
    m = pyqg.QGModel(tavestart=tmax, taveint=1, tmax=tmax, dt=dt)
    m.run()
    
    # Need to run _calc_diagnostics() once more to use the most recent state variables
    m._calc_diagnostics() 

    # Calculate spectral contribution of dissipation offline
    diss_spectrum, rhs_unfiltered = np.zeros_like(m.qh), np.zeros_like(m.qh)
    ones = np.ones_like(m.filtr)

    # Get AB coefficients
    if m.ablevel==0:
        # forward euler
        dt1 = m.dt
        dt2 = 0.0
        dt3 = 0.0
    elif m.ablevel==1:
        # AB2 at step 2
        dt1 = 1.5*m.dt
        dt2 = -0.5*m.dt
        dt3 = 0.0
    else:
        # AB3 from step 3 on
        dt1 = 23./12.*m.dt
        dt2 = -16./12.*m.dt
        dt3 = 5./12.*m.dt

    for k in range(m.nz):
        rhs_unfiltered[k] = m.qh[k] + dt1*m.dqhdt[k] + dt2*m.dqhdt_p[k] + dt3*m.dqhdt_pp[k]
        diss_spectrum[k] = (m.filtr - ones) * rhs_unfiltered[k]

    diss_contribution = -np.real(np.tensordot(m.Hi, np.conj(m.ph)*diss_spectrum, axes=(0, 0)))/m.H/m.dt/m.M**2
    diss_contribution_model = m.get_diagnostic('Dissspec')

    # Ensure that the above calculation is consistent with the model's internal calculation
    np.testing.assert_allclose(diss_contribution, diss_contribution_model, atol=atol)

    # Obtain filtered contribution, which is used in the model
    qh_new = m.qh.copy()
    for k in range(m.nz):
        qh_new[k] = m.filtr * rhs_unfiltered[k]
    rhs_contribution_filtered = -np.real(np.tensordot(m.Hi, np.conj(m.ph)*qh_new, axes=(0, 0)))/m.H/m.dt/m.M**2
    rhs_contribution_unfiltered = -np.real(np.tensordot(m.Hi, np.conj(m.ph)*rhs_unfiltered, axes=(0, 0)))/m.H/m.dt/m.M**2

    # Ensure that the difference between the filtered contribution and the unfiltered contribution is 
    # completely the effect of dissipation
    np.testing.assert_allclose(diss_contribution_model, 
                               rhs_contribution_filtered - rhs_contribution_unfiltered, 
                               atol=atol)

def test_diagnostic_magnitude():
    # Load a set of pre-run fixture models from
    # examples/diagnostic_normalization.ipynb (running from scratch would take
    # a bit too long for a test)
    fixtures_path = f"{os.path.dirname(os.path.realpath(__file__))}/fixtures"

    with open(f"{fixtures_path}/LayeredModel_params.pkl", 'rb') as f:
        # Common set of parameters for each model
        params = pickle.load(f)

    m1 = pyqg.LayeredModel(nx=96, **params)
    m2 = pyqg.LayeredModel(nx=64, **params)
    m1.q = np.load(f"{fixtures_path}/LayeredModel_nx96_q.npy")
    m2.q = np.load(f"{fixtures_path}/LayeredModel_nx64_q.npy")
    for m in [m1, m2]:
        m._invert()
        m._calc_derived_fields()

    # Loop through all diagnostics
    for diagnostic in m1.diagnostics.keys():
        if diagnostic == 'Dissspec':
            continue

        # Get the maximum-magnitude instantaneous value of each diagnostic,
        # re-evaluating the function rather than relying on any saved
        # diagnostics (gives a rough idea of order of magnitude)
        max_hi = np.abs(m1.diagnostics[diagnostic]['function'](m1)).max()
        max_lo = np.abs(m2.diagnostics[diagnostic]['function'](m2)).max()
        if max_lo == 0:
            assert max_hi == 0
        else:
            # Ensure they're the same order of magnitude -- no more than a
            # factor of 3 different. If these assertions fail for a new
            # diagnostic, you're probably missing a division by M**2.
            assert max_hi/max_lo < 3, f"{diagnostic} should be normalized"
            assert max_hi/max_lo > 0.33, f"{diagnostic} should be normalized"



"""Utility functions for pyqg model data."""

import numpy as np
from numpy import pi

def spec_var(model, ph):
    """Compute variance of ``p`` from Fourier coefficients ``ph``.

    Parameters
    ----------
    model : pyqg.Model instance
        The model object from which `ph` originates
    ph : complex array
        The field on which to compute the variance

    Returns
    -------
    var_dens : float
        The variance of `ph`
    """

    var_dens = 2. * np.abs(ph)**2 / model.M**2
    # only half of coefs [0] and [nx/2+1] due to symmetry in real fft2
    var_dens[...,0] /= 2
    var_dens[...,-1] /= 2
    return var_dens.sum(axis=(-1,-2))


def spec_sum(ph2):
    """Compute total spectral sum of the real spectral quantity``ph^2``.

    Parameters
    ----------
    model : pyqg.Model instance
        The model object from which `ph` originates
    ph2 : real array
        The field on which to compute the sum

    Returns
    -------
    var_dens : float
        The sum of `ph2`
    """

    ph2 = 2.*ph2
    ph2[...,0] = ph2[...,0]/2.
    ph2[...,-1] = ph2[...,-1]/2.

    return ph2.sum(axis=(-1,-2))

def calc_ispec(model, _var_dens, averaging = True, truncate=True, nd_wavenumber=False, nfactor = 1):
    """Compute isotropic spectrum `phr` from 2D spectrum of variable `signal2d`
    such that `signal2d.var() = phr.sum() * (kr[1] - kr[0])`.

    Parameters
    ----------
    model : pyqg.Model instance
        The model object from which `var_dens` originates
    
    var_dens : squared modulus of fourier coefficients like this:
        `np.abs(signal2d_fft)**2/m.M**2`

    averaging: If True, spectral density is estimated with averaging over circles,
        otherwise summation is used and Parseval identity holds

    truncate: If True, maximum wavenumber corresponds to inner circle in Fourier space,
        otherwise - outer circle
    
    nd_wavenumber: If True, wavenumber is nondimensional: 
        minimum wavenumber is 1 and corresponds to domain length/width,
        otherwise - wavenumber is dimensional [m^-1]

    nfactor: width of the bin in sqrt(dk^2+dl^2) units

    Returns
    -------
    kr : array
        isotropic wavenumber
    phr : array
        isotropic spectrum
    """

    # account for complex conjugate
    var_dens = np.copy(_var_dens)
    var_dens[...,0] /= 2
    var_dens[...,-1] /= 2

    ll_max = np.abs(model.ll).max()
    kk_max = np.abs(model.kk).max()

    if truncate:
        kmax = np.minimum(ll_max, kk_max)
    else:
        kmax = np.sqrt(ll_max**2 + kk_max**2)
    
    kmin = 0

    dkr = np.sqrt(model.dk**2 + model.dl**2) * nfactor

    # left border of bins
    kr = np.arange(kmin, kmax, dkr)
    
    phr = np.zeros(kr.size)

    for i in range(kr.size):
        if i == kr.size-1:
            fkr = (model.wv>=kr[i]) & (model.wv<=kr[i]+dkr)
        else:
            fkr = (model.wv>=kr[i]) & (model.wv<kr[i+1])
        if averaging:
            phr[i] = var_dens[fkr].mean() * (kr[i]+dkr/2) * pi / (model.dk * model.dl)
        else:
            phr[i] = var_dens[fkr].sum() / dkr

        phr[i] *= 2 # include full circle
    
    # convert left border of the bin to center
    kr = kr + dkr/2

    # convert to non-dimensional wavenumber 
    # preserving integral over spectrum
    if nd_wavenumber:
        kr = kr / kmin
        phr = phr * kmin

    return kr, phr



from .model import Model
from .qg_model import QGModel
from .bt_model import BTModel
from .sqg_model import SQGModel
from .layered_model import LayeredModel
from .particles import LagrangianParticleArray2D, GriddedLagrangianParticleArray2D

try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("pyqg")
except PackageNotFoundError:
    # package is not installed
    pass



import unittest
import pytest
import numpy as np
import pyqg

missing_scipy=False
try:
    import scipy.ndimage
except ImportError:
    missing_scipy=True

def constant_velocity_function(u, v):
    """Return a function that returns a constant velocity field."""
    return (lambda x, y: (u, v))

def solid_body_rotation_velocity_function(om):
    """Return a function that returns solid body rotation at angular
    velocity om."""

    def sbr(x0, y0, om0):
        r = np.sqrt(x0**2 + y0**2)
        umag = om0 * r
        a = np.arctan2(y0, x0)
        u = umag * np.sin(a)
        v = -umag * np.cos(a)
        return u, v

    return lambda x, y: sbr(x, y, om)

class ParticleTester(unittest.TestCase):

    def setUp(self):
        pass

    def test_integration(self, rtol=1e-14, atol=1e-10):
        """Check whether timestepping of particles produces the expected results."""

        x0 = np.array([1.,-1])
        y0 = np.array([0.,0.])

        # define a simple constant velocity field
        DT = 10.  # total time interval
        nt = 100  # number of time steps
        dt = DT/nt
        for u in [-1,0,1]:
            for v in [-1,0,1]:
                # compute analytical solution
                x1 = x0 + DT*u
                y1 = y0 + DT*v

                uvfun = constant_velocity_function(u, v)
                lpa = pyqg.LagrangianParticleArray2D(x0, y0)
                for n in range(nt):
                    lpa.step_forward_with_function(uvfun, uvfun, dt)

                # check solution
                np.testing.assert_allclose(x1, lpa.x, rtol=rtol)
                np.testing.assert_allclose(y1, lpa.y, rtol=rtol)

        # solid body rotation field
        nt = 1000
        for om in [0.1, 1., 10.]:
            # period, time to make a complete revolution around unit circle
            T = 2*np.pi/om
            # pick a timestep
            dt = T / nt

            uvfun = solid_body_rotation_velocity_function(om)
            lpa = pyqg.LagrangianParticleArray2D(x0, y0)
            for n in range(nt):
                lpa.step_forward_with_function(uvfun, uvfun, dt)

            # particles should be back to the same place
            np.testing.assert_allclose(
                np.array([lpa.x, lpa.y]),
                np.array([x0, y0]),
                atol=atol
            )

    @pytest.mark.skipif(missing_scipy, reason="requires scipy")
    def test_interpolation(self, rtol=1e-14, atol=1e-7):
        # set up grid
        Lx = 10.
        Ly = 5.
        Nx = 30
        Ny = 20
        dx = Lx/Nx
        dy = Ly/Ny
        xc = np.arange(Nx)*dx
        yc = np.arange(Ny)*dy
        xg = np.arange(Nx)*dx + dx/2
        yg = np.arange(Ny)*dy + dy/2
        xxg, yyg = np.meshgrid(xg, yg)
        xxc, yyc = np.meshgrid(xc, yc)

        # test field to interpolate
        ffun = lambda x, y: np.sin(2*np.pi*x/Lx)*np.sin(2*np.pi*y/Ly)
        f_at_g = ffun(xxg, yyg)
        f_at_c = ffun(xxc, yyc)

        Npart = 10
        x0 = np.array(np.random.rand(Npart)*Lx)
        y0 = np.array(np.random.rand(Npart)*Ly)
        glpa = pyqg.GriddedLagrangianParticleArray2D(
             x0, y0, Nx, Ny,
             xmin=0, xmax=Lx, ymin=0, ymax=Ly,
             periodic_in_x=True, periodic_in_y=True)

        # check if we interpolate back to the same points, we get same result
        f_at_gi = glpa.interpolate_gridded_scalar(xxg, yyg, f_at_g)
        np.testing.assert_allclose(f_at_g, f_at_gi)

        # now try shifting everything
        # used to use cubic interpolation by default
        # that was slow and produced weird artifacts near boundaries
        # now use linear interpolation
        # this definitely produces errors in the approximation of sine and cosine
        # neglect of curvature causes velocity to be systematically underestimated
        f_at_ci = glpa.interpolate_gridded_scalar(xxc, yyc, f_at_g)
        np.testing.assert_allclose(f_at_c, f_at_ci, atol=1e-1)

        # now try some random points
        # what sort of error do we expect
        ci = ffun(x0, y0)
        np.testing.assert_allclose(ci,
            glpa.interpolate_gridded_scalar(x0, y0, f_at_g),
            atol=1e-1
        )

    @pytest.mark.skipif(missing_scipy, reason="requires scipy")
    def test_gridded_integration(self, atol=1e-10):

        # set up grid
        Lx = 4.
        Ly = 4.
        Nx = 100
        Ny = 100
        dx = Lx/Nx
        dy = Ly/Ny
        # corner points
        xc = np.arange(Nx)*dx - Lx/2
        yc = np.arange(Ny)*dy - Lx/2
        # center points
        xg = xc + dx/2
        yg = xc + dy/2
        xxg, yyg = np.meshgrid(xg, yg)
        xxc, yyc = np.meshgrid(xc, yc)

        # initial particle positions
        x0 = np.array([1.,-1])
        y0 = np.array([0.,0.])

        # solid body rotation field
        nt = 1000
        for om in [0.1, 1., 10.]:
            # period, time to make a complete revolution around unit circle
            T = 2*np.pi/om
            # pick a timestep
            dt = T / nt

            uvfun = solid_body_rotation_velocity_function(om)
            u_at_g, v_at_g = uvfun(xxg, yyg)
            glpa = pyqg.GriddedLagrangianParticleArray2D(
                 x0, y0, Nx, Ny,
                 xmin=-Lx/2, xmax=Lx/2, ymin=-Ly/2, ymax=Ly/2,
                 periodic_in_x=True, periodic_in_y=True)

            for n in range(nt):
                glpa.step_forward_with_gridded_uv(
                    u_at_g, v_at_g, u_at_g, v_at_g, dt)

            # particles should be back to the same place
            np.testing.assert_allclose(
                np.array([glpa.x, glpa.y]),
                np.array([x0, y0]),
                atol=atol
            )






if __name__ == '__main__':
    unittest.main()



import numpy as np
import pyqg

def test_advect(rtol=1.e-13):
    """ Make sure advective term vanishes for plane wave
        It is an unpleasant fact that we cannot to get
        a double precision accuracy when the plane wave is
        slanted (kx != 0 and ky !=0 ) """

    #m = bt_model.BTModel(L=2.*np.pi,nx=256)
    m = pyqg.QGModel(L=2.*np.pi,nx=256,U1=0)

    # there are some magic combinations that
    #   fails the test...try kx=12,ky=24
    #   should investigate what's going on...

    kx = np.array([1.,5.,10.,0.,24.,2.])
    ky = np.array([2.,0.,10.,21.,12.,49.])


    for i in range(kx.size):

        # set plane wave PV
        #m.set_q(
        #        np.cos( kx[i] * m.x + ky[i] * m.y ))

        m.set_q1q2(
                np.cos( kx[i] * m.x + ky[i] * m.y ),
                np.zeros_like(m.x) )

        # compute psi
        m._invert()
        #m.ph1,m.ph2 = m.invph(m.qh1,m.qh2)

        # diagnose vel.
        #m.u1,m.v1 = m.caluv(m.ph1)

        # compute advection
        #jacobh = m.advect(m.q[0],m.u[0],m.v[0])
        jacobh = m._advect(m.q,m.u,m.v)
        #jacobh = m.advect(m.q,m.u,m.v)
        jacob = m.ifft(jacobh)

        # residual -- the L2-norm of the jacobian
        res = np.abs(jacob).sum()*m.dx*m.dy/(m.L**2)

        print("residual = %1.5e" %res)

        assert res<rtol, " *** Jacobian residual is larger than %1.1e" %rtol

if __name__ == "__main__":
    test_advect()



import matplotlib.pyplot as plt
import numpy as np
from numpy import pi
import pyqg

# the model object
year = 1.
m = pyqg.BTModel(L=2.*pi,nx=256, tmax = 50*year,
        beta = 21., H = 1., rek = 0., rd = None, dt = 0.001,
                     taveint=year, ntd=4)

fk = m.wv != 0
ckappa = np.zeros_like(m.wv2)
ckappa[fk] = np.sqrt( m.wv2[fk]*(1. + (m.wv2[fk]/36.)**2) )**-1

nhx,nhy = m.wv2.shape

Pi_hat = np.random.randn(nhx,nhy)*ckappa +1j*np.random.randn(nhx,nhy)*ckappa

Pi = m.ifft( Pi_hat[np.newaxis,:,:] )
Pi = Pi - Pi.mean()

Pi = Pi - Pi.mean()
Pi_hat = m.fft( Pi )
KEaux = m.spec_var( m.wv*Pi_hat )

pih = ( Pi_hat/np.sqrt(KEaux) )
qih = -m.wv2*pih
qi = m.ifft(qih)
m.set_q(qi)

# run the model
plt.rcParams['image.cmap'] = 'YlOrRd_r'

for snapshot in m.run_with_snapshots(tsnapstart=0, tsnapint=200*m.dt):
    plt.clf()
    plt.imshow(m.q.squeeze() + m.beta * m.y)
    plt.clim([-20., 130.])
    #plt.title(str(m.t))
    plt.pause(0.01)

    plt.draw()




import unittest
import numpy as np
from numpy import sqrt
import pyqg

class LinearStabilityTester(unittest.TestCase):

    def setUp(self):

        self.atol=1.e-16

    def test_two_layer_stability(self):
        """ Make growth rates calculates numerically
                agree with the exact dispersion relationship
                for nz = 2 """

        m = pyqg.LayeredModel(L=1.e6,rd = 15.e3,nx=256,U=np.array([.1,0.]),V=np.array([0.,0.]),
                H=np.array([2000,2000.]),delta=1.,nz=2,f=1.)

        # numerical results
        evals, evecs = m.stability_analysis()

        # analytical results
        kb = sqrt(m.beta/(m.Us/2.))
        wv4 = m.wv2**2
        kd2 = m.rd**-2
        kd4 = kd2**2
        kb4 = kb**4
        omg_ana = np.zeros_like(m.wv2) + 0.j
        D = 1. +  (4.*wv4*(wv4 - kd4))/(kb4*kd4)

        fneg = D<0.
        omg_ana[fneg] = 1j*m.k[fneg]*(m.beta/(m.wv2[fneg] + kd2))*( (kd2/(2.*m.wv2[fneg]))*sqrt(-D[fneg]) )

        res = np.abs(omg_ana.imag-evals.imag).max()

        np.testing.assert_allclose(res,0.,atol=self.atol,
                err_msg=' Residual  of growth rate is larger than %1.1e' %res)

if __name__ == "__main__":
    unittest.main()



import matplotlib.pyplot as plt
import numpy as np
from numpy import pi
import pyqg
from pyqg.diagnostic_tools import spec_var

# the model object
year = 1.
m = pyqg.BTModel(L=2.*pi,nx=128, tmax = 200*year,
        beta = 0., H = 1., rek = 0., rd = None, dt = 0.005,
                     taveint=year, ntd=2)

# McWilliams 84 IC condition
fk = m.wv != 0
ckappa = np.zeros_like(m.wv2)
ckappa[fk] = np.sqrt( m.wv2[fk]*(1. + (m.wv2[fk]/36.)**2) )**-1

nhx,nhy = m.wv2.shape

Pi_hat = np.random.randn(nhx,nhy)*ckappa +1j*np.random.randn(nhx,nhy)*ckappa

Pi = m.ifft( Pi_hat[np.newaxis] )
Pi = Pi - Pi.mean()
Pi_hat = m.fft( Pi )
KEaux = spec_var( m, m.filtr*m.wv*Pi_hat )

pih = ( Pi_hat/np.sqrt(KEaux) )
qih = -m.wv2*pih
qi = m.ifft(qih)
m.set_q(qi)

# run the model and plot some figs
plt.rcParams['image.cmap'] = 'RdBu_r'

plt.ion()

for snapshot in m.run_with_snapshots(tsnapstart=0, tsnapint=15*m.dt):

    plt.clf()
    p1 = plt.imshow(m.q[0] + m.beta * m.y)
    plt.clim([-30., 30.])
    plt.title('t='+str(m.t))
    
    plt.xticks([])
    plt.yticks([])

    plt.pause(0.01)

    plt.draw()

plt.show()
plt.ion()



import numpy as np
from matplotlib import pyplot as plt
import pyqg

m = pyqg.QGModel(tavestart=0,  dt=8000)

for snapshot in m.run_with_snapshots(
        tsnapstart=0, tsnapint=1000*m.dt):
    plt.clf()
    plt.imshow(m.q[0] + m.Qy1 * m.y)
    plt.clim([0,  m.Qy1 * m.W])
    plt.pause(0.01)
    plt.draw()
    
# now the model is done



import numpy as np
from numpy import pi
import logging
import warnings

from .errors import DiagnosticNotFilledError
from .kernel import PseudoSpectralKernel, tendency_forward_euler, tendency_ab2, tendency_ab3
try:
    import mkl
    np.use_fastnumpy = True
except ImportError:
    pass

try:
    import pyfftw
    pyfftw.interfaces.cache.enable()
except ImportError:
    pass

class Model(PseudoSpectralKernel):
    """A generic pseudo-spectral inversion model.

    Attributes
    ----------
    nx, ny : int
        Number of real space grid points in the `x`, `y` directions (cython)
    nk, nl : int
        Number of spectral space grid points in the `k`, `l` directions (cython)
    nz : int
        Number of vertical levels (cython)
    kk, ll : real array
        Zonal and meridional wavenumbers (`nk`) (cython)
    a : real array
        inversion matrix (`nk`, `nk`, `nl`, `nk`) (cython)
    q : real array
        Potential vorticity in real space (`nz`, `ny`, `nx`) (cython)
    qh : complex array
        Potential vorticity in spectral space (`nk`, `nl`, `nk`) (cython)
    ph : complex array
        Streamfunction in spectral space (`nk`, `nl`, `nk`) (cython)
    u, v : real array
        Zonal and meridional velocity anomalies in real space (`nz`, `ny`, `nx`) (cython)
    Ubg : real array
        Background zonal velocity (`nk`) (cython)
    Qy : real array
        Background potential vorticity gradient (`nk`) (cython)
    ufull, vfull : real arrays
        Zonal and meridional full velocities in real space (`nz`, `ny`, `nx`) (cython)
    uh, vh : complex arrays
        Velocity anomaly components in spectral space (`nk`, `nl`, `nk`) (cython)
    rek : float
        Linear drag in lower layer (cython)
    t : float
        Model time (cython)
    tc : int
        Model timestep (cython)
    dt : float
        Numerical timestep (cython)
    L, W : float
        Domain length in x and y directions
    filterfac : float
        Amplitdue of the spectral spherical filter
    twrite : int
        Interval for cfl writeout (units: number of timesteps)
    tmax : float
        Total time of integration (units: model time)
    tavestart : float
        Start time for averaging (units: model time)
    tsnapstart : float
        Start time for snapshot writeout (units: model time)
    taveint : float
        Time interval for accumulation of diagnostic averages.
        (units: model time)
    tsnapint : float
        Time interval for snapshots (units: model time)
    ntd : int
        Number of threads to use. Should not exceed the number of cores on
        your machine.
    pmodes : real array
        Vertical pressure modes (unitless)
    radii :  real array
        Deformation radii  (units: model length)
    q_parameterization : function
        Optional function which takes the model as input and returns a `numpy`
        array of shape (`nz`, `ny`, `nx`) to be added to dq/dt before stepping
        forward. This can be used to implement subgrid forcing
        parameterizations.
    uv_parameterization : function
        Optional function which takes the model as input and returns a tuple of
        two `numpy` arrays, each of shape (`nz`, `ny`, `nx`), to be added to
        the zonal and meridional velocity derivatives (respectively) at each
        timestep. This can also be used to implemented subgrid forcing
        parameterizations, but expressed in terms of velocity rather than
        potential vorticity.
    """

    def __init__(
        self,
        # grid size parameters
        nz=1,
        nx=64,                     # grid resolution
        ny=None,
        L=1e6,                     # domain size is L [m]
        W=None,
        # timestepping parameters
        dt=7200.,                   # numerical timestep
        twrite=1000.,               # interval for cfl and ke writeout (in timesteps)
        tmax=1576800000.,           # total time of integration
        tavestart=315360000.,       # start time for averaging
        taveint=86400.,             # time interval used for summation in longterm average in seconds
        useAB2=False,               # use second order Adams Bashforth timestepping instead of 3rd
        # friction parameters
        rek=5.787e-7,               # linear drag in lower layer
        filterfac=23.6,             # the factor for use in the exponential filter
        # constants
        f = None,                   # coriolis parameter (not necessary for two-layer model
                                    #  if deformation radius is provided)
        g= 9.81,                    # acceleration due to gravity
        q_parameterization=None,    # subgrid parameterization in terms of q
        uv_parameterization=None,   # subgrid parameterization in terms of u,v
        # diagnostics parameters
        diagnostics_list='all',     # which diagnostics to output
        # fft parameters
        # removed because fftw is now manditory
        #use_fftw = False,               # fftw flag
        #teststyle = False,            # use fftw with "estimate" planner to get reproducibility
        ntd = 1,                       # number of threads to use in fftw computations
        log_level = 1,                 # logger level: from 0 for quiet (no log) to 4 for verbose
                                       #     logger (see  https://docs.python.org/2/library/logging.html)
        logfile = None,                # logfile; None prints to screen
        ):
        """
        .. note:: All of the test cases use ``nx==ny``. Expect bugs if you choose
                  these parameters to be different.
        .. note:: All time intervals will be rounded to nearest `dt` interval.

        Parameters
        ----------
        nx : int
            Number of grid points in the x direction.
        ny : int
            Number of grid points in the y direction (default: nx).
        L : number
            Domain length in x direction. Units: meters.
        W :
            Domain width in y direction. Units: meters (default: L).
        rek : number
            linear drag in lower layer. Units: seconds :sup:`-1`.
        filterfac : number
            amplitdue of the spectral spherical filter (originally 18.4, later
            changed to 23.6).
        dt : number
            Numerical timstep. Units: seconds.
        twrite : int
            Interval for cfl writeout. Units: number of timesteps.
        tmax : number
            Total time of integration. Units: seconds.
        tavestart : number
            Start time for averaging. Units: seconds.
        tsnapstart : number
            Start time for snapshot writeout. Units: seconds.
        taveint : number
            Time interval for accumulation of diagnostic averages.
            Units: seconds. (For performance purposes, averaging does not have to
            occur every timestep)
        tsnapint : number
            Time interval for snapshots. Units: seconds.
        ntd : int
            Number of threads to use. Should not exceed the number of cores on
            your machine.
        q_parameterization : function
            Optional function which takes the model as input and returns a `numpy`
            array of shape (`nz`, `ny`, `nx`) to be added to dq/dt before stepping
            forward. This can be used to implement subgrid forcing
            parameterizations.
        uv_parameterization : function
            Optional function which takes the model as input and returns a tuple of
            two `numpy` arrays, each of shape (`nz`, `ny`, `nx`), to be added to
            the zonal and meridional velocity derivatives (respectively) at each
            timestep. This can also be used to implemented subgrid forcing
            parameterizations, but expressed in terms of velocity rather than
            potential vorticity.
        """

        if ny is None:
            ny = nx
        if W is None:
            W = L

        # TODO: be more clear about what attributes are cython and what
        # attributes are python
        PseudoSpectralKernel.__init__(self, nz, ny, nx, ntd,
                has_q_param=int(q_parameterization is not None),
                has_uv_param=int(uv_parameterization is not None))

        self.L = L
        self.W = W

        # timestepping
        self.dt = dt
        self.twrite = twrite
        self.tmax = tmax
        self.tavestart = tavestart
        self.taveint = taveint
        self.logfile = logfile
        self.log_level = log_level
        self.useAB2 = useAB2
        self.ntd = ntd

        # friction
        self.rek = rek
        self.filterfac = filterfac

        # constants
        self.g = g
        if f:
            self.f = f
            self.f2 = f**2

        # optional subgrid parameterizations
        self.q_parameterization = q_parameterization
        self.uv_parameterization = uv_parameterization

        # TODO: make this less complicated!
        # Really we just need to initialize the grid here. It's not necessary
        # to have all these silly methods. Maybe we need "hooks" instead.
        self._initialize_logger()
        self._initialize_grid()
        self._initialize_background()
        self._initialize_forcing()
        self._initialize_filter()
        self._initialize_time()
        self._initialize_inversion_matrix()
        self._initialize_diagnostics(diagnostics_list)


    def run_with_snapshots(self, tsnapstart=0., tsnapint=432000.):
        """Run the model forward, yielding to user code at specified intervals.

        Parameters
        ----------

        tsnapstart : int
            The timestep at which to begin yielding.
        tstapint : int
            The interval at which to yield.
        """

        tsnapints = np.ceil(tsnapint/self.dt)

        while(self.t < self.tmax):
            self._step_forward()
            if self.t>=tsnapstart and (self.tc%tsnapints)==0:
                yield self.t
        return

    def run(self):
        """Run the model forward without stopping until the end."""
        while(self.t < self.tmax):
            self._step_forward()


    def vertical_modes(self):
        """ Calculate standard vertical modes. Simply
            the eigenvectors of the stretching matrix S """

        evals,evecs = np.linalg.eig(-self.S)

        asort = evals.argsort()

        # deformation wavenumbers and radii
        self.kdi2 = evals[asort]
        self.radii = np.zeros_like(self.kdi2)
        self.radii[0] = np.sqrt(self.g*self.H)/np.abs(self.f) # barotropic def. radius
        self.radii[1:] = 1./np.sqrt(self.kdi2[1:])

        # eigenstructure
        self.pmodes = evecs[:,asort]

        # normalize to have unit L2-norm
        Ai = (self.H / (self.Hi[:,np.newaxis]*(self.pmodes**2)).sum(axis=0))**0.5
        self.pmodes = Ai[np.newaxis,:]*self.pmodes

    def modal_projection(self,p,forward=True):
        """ Performs a field p into modal amplitudes pn
                using the basis [pmodes]. The inverse
                transform calculates p from pn"""

        if forward:
            pt = np.linalg.solve(self.pmodes[np.newaxis,np.newaxis],p.T).T
        else:
            pt = np.einsum("ik,k...->i...",self.pmodes,p)

        return pt

    def stability_analysis(self,bottom_friction=False):
        r""" Performs the baroclinic linear instability analysis given
                given the base state velocity :math: `(U, V)` and
                the stretching matrix  :math: `S`:

        .. math:: A \Phi = \omega B \Phi,

        where

        .. math:: A = B (U k + V l) + I (k Q_y - l Q_x) +
                        1j \delta_{N N} r_{ek} I  \kappa^2

        where :math:`\delta_{N N} = [0,0,\dots,0,1] ,`

        and

        .. math:: B =  S - I \kappa^2 .


        The eigenstructure is

        .. math:: \Phi

        and the eigenvalue is

        .. math:: `\omega`

        The growth rate is Im\ :math:`\{\omega\}`.


        Parameters
        ----------
        bottom_friction: optional inclusion linear bottom drag
                         in the linear stability calculation
                         (default is False, as if :math: `r_{ek} = 0`)

        Returns
        -------
        omega: complex array
             The eigenvalues with largest complex part (units: inverse model time)
        phi: complex array
             The eigenvectors associated associated with \omega (unitless)

        """

        omega = np.zeros_like(self.wv)+0.j
        phi = np.zeros_like(self.qh)

        I = np.eye(self.nz)

        L2 = self.S[:,:,np.newaxis,np.newaxis] - self.wv2*I[:,:,np.newaxis,np.newaxis]

        Q =  I[:,:,np.newaxis,np.newaxis]*(self.ikQy - self.ilQx).imag

        Uk =(self.Ubg*I)[:,:,np.newaxis,np.newaxis]*self.k
        Vl =(self.Vbg*I)[:,:,np.newaxis,np.newaxis]*self.l
        L3 = np.einsum('ij...,jk...->ik...',L2,Uk+Vl) + 0j

        if bottom_friction:
            L3[-1,-1,:,:] += 1j*self.rek*self.wv2

        L4 = self.a.T

        M = np.einsum('...ij,...jk->...ik',L4,(L3+Q).T)

        evals,evecs = np.linalg.eig(M)
        evals, evecs = evals.T, evecs.T

        # sorting things this way proved way
        #  more faster than using numpy's argsort() !
        imax = evals.imag.argmax(axis=0)
        for i in range(self.nl):
            for j in range(self.nk):
                omega[i,j] = evals[imax[i,j],i,j]
                phi[:,i,j] = evecs[imax[i,j],:,i,j]

        return omega, phi

    ### PRIVATE METHODS - not meant to be called by user ###

    def _step_forward(self):

        self._invert()
        # find streamfunction from pv

        self._do_advection()
        # use streamfunction to calculate advection tendency

        self._do_friction()
        # apply friction

        self._do_external_forcing()
        # apply external forcing

        if self.uv_parameterization is not None:
            self._do_uv_subgrid_parameterization()
            # apply velocity subgrid forcing term, if present

        if self.q_parameterization is not None:
            self._do_q_subgrid_parameterization()
            # apply potential vorticity subgrid forcing term, if present

        self._calc_diagnostics()
        # do what has to be done with diagnostics

        self._forward_timestep()
        # apply tendencies to step the model forward
        # (filter gets called here)

        # the basic steps are
        self._print_status()

    def _initialize_time(self):
        """Set up timestep stuff"""
        #self.t=0        # actual time
        #self.tc=0       # timestep number
        self.taveints = np.ceil(self.taveint/self.dt)

    ### initialization routines, only called once at the beginning ###
    # TODO: clean up and simplify this whole routine
    def _initialize_grid(self):
        """Set up spatial and spectral grids and related constants"""
        self.x,self.y = np.meshgrid(
            np.arange(0.5,self.nx,1.)/self.nx*self.L,
            np.arange(0.5,self.ny,1.)/self.ny*self.W )

        # Notice: at xi=1 U=beta*rd^2 = c for xi>1 => U>c
        # wavenumber one (equals to dkx/dky)
        self.dk = 2.*pi/self.L
        self.dl = 2.*pi/self.W

        # wavenumber grids
        # set in kernel
        #self.nl = self.ny
        #self.nk = int(self.nx/2+1)
        self.ll = self.dl*np.append( np.arange(0.,self.nx/2),
            np.arange(-self.nx/2,0.) )
        self.kk = self.dk*np.arange(0.,self.nk)

        self.k, self.l = np.meshgrid(self.kk, self.ll)
        self.ik = 1j*self.k
        self.il = 1j*self.l
        # physical grid spacing
        self.dx = self.L / self.nx
        self.dy = self.W / self.ny

        # constant for spectral normalizations
        self.M = self.nx*self.ny

        # isotropic wavenumber^2 grid
        # the inversion is not defined at kappa = 0
        self.wv2 = self.k**2 + self.l**2
        self.wv = np.sqrt( self.wv2 )

        iwv2 = self.wv2 != 0.
        self.wv2i = np.zeros_like(self.wv2)
        self.wv2i[iwv2] = self.wv2[iwv2]**-1

    def _initialize_background(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_inversion_matrix(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_forcing(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_filter(self):
        """Set up frictional filter."""
        # this defines the spectral filter (following Arbic and Flierl, 2003)
        cphi=0.65*pi
        wvx=np.sqrt((self.k*self.dx)**2.+(self.l*self.dy)**2.)
        filtr = np.exp(-self.filterfac*(wvx-cphi)**4.)
        filtr[wvx<=cphi] = 1.
        self.filtr = filtr

    def _filter(self, q):
        return self.filtr * q

    def _do_external_forcing(self):
        pass

    # logger
    def _initialize_logger(self):

        self.logger = logging.getLogger(__name__)


        if not (self.logfile is None):
            fhandler = logging.FileHandler(filename=self.logfile, mode='w')
        else:
            fhandler = logging.StreamHandler()

        formatter = logging.Formatter('%(levelname)s: %(message)s')

        fhandler.setFormatter(formatter)

        if not self.logger.handlers:
            self.logger.addHandler(fhandler)

        self.logger.setLevel(self.log_level*10)

        # this prevents the logger to propagate into the ipython notebook log
        self.logger.propagate = False

        self.logger.info(' Logger initialized')


    # compute advection in grid space (returns qdot in fourier space)
    # *** don't remove! needed for diagnostics (but not forward model) ***
    def _advect(self, q, u, v):
        """Given real inputs q, u, v, returns the advective tendency for
        q in spectral space."""
        uq = u*q
        vq = v*q
        # this is a hack, since fft now requires input to have shape (nz,ny,nx)
        # it does an extra unnecessary fft
        is_2d = (uq.ndim==2)
        if is_2d:
            uq = np.tile(uq[np.newaxis,:,:], (self.nz,1,1))
            vq = np.tile(vq[np.newaxis,:,:], (self.nz,1,1))
        tend = self.ik*self.fft(uq) + self.il*self.fft(vq)
        if is_2d:
            return tend[0]
        else:
            return tend

    # def _filter(self, q):
    #     """Apply filter to field q."""
    #     return q

    def _print_status(self):
        """Output some basic stats."""
        if (self.log_level) and ((self.tc % self.twrite)==0):
            self.ke = self._calc_ke()
            self.cfl = self._calc_cfl()
            #print 't=%16d, tc=%10d: cfl=%5.6f, ke=%9.9f' % (
            #       self.t, self.tc, cfl, ke)
            self.logger.info('Step: %i, Time: %3.2e, KE: %3.2e, CFL: %4.3f'
                    , self.tc,self.t,self.ke,self.cfl )

            assert self.cfl<1., self.logger.error('CFL condition violated')


    def _calc_diagnostics(self):
        # here is where we calculate diagnostics
        if (self.t>=self.dt) and (self.t>=self.tavestart) and (self.tc%self.taveints==0):
            self._increment_diagnostics()

    # def _forward_timestep(self):
    #     """Step forward based on tendencies"""
    #
    #     #self.dqhdt = self.dqhdt_adv + self.dqhdt_forc
    #
    #     # Note that Adams-Bashforth is not self-starting
    #     if self.tc==0:
    #         # forward Euler at the first step
    #         qtend = tendency_forward_euler(self.dt, self.dqhdt)
    #     elif (self.tc==1) or (self.useAB2):
    #         # AB2 at step 2
    #         qtend = tendency_ab2(self.dt, self.dqhdt, self.dqhdt_p)
    #     else:
    #         # AB3 from step 3 on
    #         qtend = tendency_ab3(self.dt, self.dqhdt,
    #                     self.dqhdt_p, self.dqhdt_pp)
    #
    #     # add tendency and filter
    #     self.set_qh(self._filter(self.qh + qtend))
    #
    #     # remember previous tendencies
    #     self.dqhdt_pp[:] = self.dqhdt_p.copy()
    #     self.dqhdt_p[:] = self.dqhdt.copy()
    #     #self.dqhdt[:] = 0.
    #
    #     # augment timestep and current time
    #     self.tc += 1
    #     self.t += self.dt

    ### All the diagnostic stuff follows. ###
    def _calc_cfl(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    # this is stuff the Cesar added

    # if self.tc==0:
    #     assert self.calc_cfl()<1., " *** time-step too large "
    #     # initialize ke and time arrays
    #     self.ke = np.array([self.calc_ke()])
    #     self.eddy_time = np.array([self.calc_eddy_time()])
    #     self.time = np.array([0.])


    def _calc_ke(self):
        raise NotImplementedError(
            'needs to be implemented by Model subclass')

    def _initialize_diagnostics(self, diagnostics_list):
        # Initialization for diagnotics
        self.diagnostics = dict()

        self._initialize_core_diagnostics()
        self._initialize_model_diagnostics()

        if diagnostics_list == 'all':
            pass # by default, all diagnostics are active
        elif diagnostics_list == 'none':
            self.set_active_diagnostics([])
        else:
            self.set_active_diagnostics(diagnostics_list)

    def _initialize_core_diagnostics(self):
        """Diagnostics common to all models."""
       
        self.add_diagnostic('Ensspec',
            description='enstrophy spectrum',
            function= (lambda self: np.abs(self.qh)**2/self.M**2),
            units='s^-2',
            dims=('lev','l','k')         
        )

        self.add_diagnostic('KEspec',
            description='kinetic energy spectrum',
            function= (lambda self: self.wv2*np.abs(self.ph)**2/self.M**2),
            units='m^2 s^-2',
            dims=('lev','l','k')  
        )      # factor of 2 to account for the fact that we have only half of
               #    the Fourier coefficients.

        self.add_diagnostic('EKEdiss',
            description='total energy dissipation by bottom drag',
            function= (lambda self: self.Hi[-1]/self.H*self.rek*(self.v[-1]**2 + self.u[-1]**2).mean()),
            units='m^2 s^-3',
            dims=('time',)
        )

        self.add_diagnostic('KEfrictionspec',
            description='total energy dissipation spectrum by bottom drag',
            function= (lambda self: -self.rek*self.Hi[-1]/self.H*self.wv2*np.abs(self.ph[-1])**2/self.M**2),
            units='m^2 s^-3',
            dims=('l','k')
        )

        self.add_diagnostic('EKE',
            description='mean eddy kinetic energy',
            function= (lambda self: 0.5*(self.v**2 + self.u**2).mean(axis=-1).mean(axis=-1)),
            units='m^2 s^-2',
            dims=('lev',)
        )

        def dissipation_spectrum(m):
            spectrum = np.zeros_like(m.qh)
            ones = np.ones_like(m.filtr)
            if m.ablevel==0:
                # forward euler
                dt1 = m.dt
                dt2 = 0.0
                dt3 = 0.0
            elif m.ablevel==1:
                # AB2 at step 2
                dt1 = 1.5*m.dt
                dt2 = -0.5*m.dt
                dt3 = 0.0
            else:
                # AB3 from step 3 on
                dt1 = 23./12.*m.dt
                dt2 = -16./12.*m.dt
                dt3 = 5./12.*m.dt
            for k in range(m.nz):
                spectrum[k] = (m.filtr - ones) * (
                    m.qh[k] + dt1*m.dqhdt[k] + dt2*m.dqhdt_p[k] + dt3*m.dqhdt_pp[k])
            return spectrum

        self.add_diagnostic('Dissspec',
            description='Spectral contribution of filter dissipation to total energy',
            function=(lambda self: -np.tensordot(self.Hi, 
                np.conj(self.ph)*dissipation_spectrum(self), axes=(0, 0)).real/self.H/self.dt/self.M**2),
            units='m^2 s^-3',
            dims=('l','k')
        )

        self.add_diagnostic('ENSDissspec',
            description='Spectral contribution of filter dissipation to barotropic enstrophy',
            function=(lambda self: np.tensordot(self.Hi, 
                np.conj(self.qh)*dissipation_spectrum(self), axes=(0, 0)).real/self.H/self.dt/self.M**2),
            units='s^-3',
            dims=('l','k')
        )

        self.add_diagnostic('paramspec',
            description='Spectral contribution of subgrid parameterization (if present)',
            function=lambda self: self._calc_parameterization_spectrum(),
            units='m^2 s^-3',
            dims=('l','k')
        )

    def _calc_parameterization_contribution(self):
        dqh = np.zeros_like(self.qh)
        if self.uv_parameterization is not None:
            ik = np.asarray(self._ik).reshape((1, -1)).repeat(self.wv2.shape[0], axis=0)
            il = np.asarray(self._il).reshape((-1, 1)).repeat(self.wv2.shape[-1], axis=-1)
            dqh += -il * self.duh + ik * self.dvh
        if self.q_parameterization is not None:
            dqh += self.dqh
        return dqh

    def _calc_parameterization_spectrum(self):
        dqh = self._calc_parameterization_contribution()
        height_ratios = (self.Hi / self.H)[:,np.newaxis,np.newaxis]
        return -np.real((height_ratios * np.conj(self.ph) * dqh).sum(axis=0)) / self.M**2

    def _calc_derived_fields(self):
        """Should be implemented by subclass."""
        pass

    def _initialize_model_diagnostics(self):
        """Should be implemented by subclass."""
        pass

    def _set_active_diagnostics(self, diagnostics_list):
        for d in self.diagnostics:
            self.diagnostics[d]['active'] == (d in diagnostics_list)

    def add_diagnostic(self, diag_name, description=None, function=None, units=None, dims=None):
        # create a new diagnostic dict and add it to the object array

        # make sure the function is callable
        assert hasattr(function, '__call__')

        # make sure the name is valid
        assert isinstance(diag_name, str)

        # by default, diagnostic is active
        self.diagnostics[diag_name] = {
           'description': description,
           'units': units,
           'dims': dims,
           'active': True,
           'count': 0,
           'function': function, }

    def describe_diagnostics(self):
        """Print a human-readable summary of the available diagnostics."""
        diag_names = list(self.diagnostics.keys())
        diag_names.sort()
        print('NAME               | DESCRIPTION')
        print(80*'-')
        for k in diag_names:
            d = self.diagnostics[k]
            print('{:<10} | {:<54}'.format(
                 *(k,  d['description'])))

    def _increment_diagnostics(self):
        # compute intermediate quantities needed for some diagnostics

        self._calc_derived_fields()

        for dname in self.diagnostics:
            if self.diagnostics[dname]['active']:
                res = self.diagnostics[dname]['function'](self)
                if self.diagnostics[dname]['count']==0:
                    self.diagnostics[dname]['value'] = res
                else:
                    self.diagnostics[dname]['value'] += res
                self.diagnostics[dname]['count'] += 1

    def get_diagnostic(self, dname):
        if 'value' not in self.diagnostics[dname]:
            raise DiagnosticNotFilledError(dname)
        return (self.diagnostics[dname]['value'] /
                self.diagnostics[dname]['count'])

    def spec_var(self, ph):
        """ compute variance of p from Fourier coefficients ph """
        var_dens = 2. * np.abs(ph)**2 / self.M**2
        # only half of coefs [0] and [nx/2+1] due to symmetry in real fft2
        var_dens[...,0] = var_dens[...,0]/2.
        var_dens[...,-1] = var_dens[...,-1]/2.
        return var_dens.sum()

    def set_qh(self, qh):
        warnings.warn("Method deprecated. Set model.qh directly instead. ",
            DeprecationWarning)
        self.qh = qh

    def set_q(self, q):
        warnings.warn("Method deprecated. Set model.q directly instead. ",
            DeprecationWarning)
        self.q = q

    def to_dataset(self):
        """Convert outputs from model to an xarray dataset
        
        Returns
        -------
        ds : xarray.Dataset
        """
        from .xarray_output import model_to_dataset
        return model_to_dataset(self)



class PYQGError(Exception):
    """Base class for all PYQG exceptions"""
    pass

class DiagnosticNotFilledError(PYQGError):
    """Raise this when trying access a diagnostic that has not been filled"""
    def __init__(self, dname):
        message = f"Diagnostic {dname} has not been filled. Please adjust tavestart and taveint."
        super().__init__(message)
    pass


import matplotlib.pyplot as plt
import numpy as np
from numpy import pi
from pyqg import sqg_model

# the model object
year = 1.
m = sqg_model.SQGModel(L=2.*pi,nx=512, tmax = 40.005,
        beta = 0., Nb = 1., H = 1., f_0 = 1, dt = 0.005,
                     taveint=1, ntd=4)

# run the model and plot some figs
plt.rcParams['image.cmap'] = 'RdBu'

# Choose ICs from Held et al. (1996)
# case i) Elliptical vortex
x = np.linspace(m.dx/2,2*np.pi,m.nx) - np.pi
y = np.linspace(m.dy/2,2*np.pi,m.ny) - np.pi
x,y = np.meshgrid(x,y)

qi = -np.exp(-(x**2 + (4.0*y)**2)/(m.L/6.0)**2)
m.set_q(qi[np.newaxis,:,:])

plt.ion()

for snapshot in m.run_with_snapshots(tsnapstart=0, tsnapint=200*m.dt):
    plt.clf()
    p1 = plt.imshow(m.q.squeeze() + m.beta * m.y)
    #plt.clim([-30., 30.])
    plt.title('t='+str(m.t))
    plt.colorbar()
    plt.clim([-1, 0])
    plt.xticks([])
    plt.yticks([])
    plt.draw()

    plt.pause(0.01)


plt.ioff()
plt.show()



import numpy as np
import unittest
import pyqg
from pyqg.diagnostic_tools import spec_var

class ModelFFTMixin:

    def test_parseval(self, rtol=1.e-15):
        """ Make sure 2D fft satisfies Parseval's relation
                within machine double precision """

        # set random stream function
        p1 = np.random.randn(self.m.nz, self.m.ny, self.m.nx)
        p1 -= p1.mean(axis=(-2,-1))[:,np.newaxis,np.newaxis]

        ph1 = self.m.fft(p1)

        # var(P) from Fourier coefficients
        P_var_spec = spec_var(self.m, ph1)
        #print("Variance from spectrum: %5.16f" % P_var_spec)

        # var(P) in physical space
        P_var_phys = p1.var(axis=(-2,-1))
        #print("Variance in physical space: %5.16f" % P_var_phys)

        # relative error
        error = np.abs((P_var_phys - P_var_spec)/P_var_phys).sum()
        print("error = %5.16f" %error)

        self.assertTrue(error<rtol,
                "fft does not satisfy Parseval's relation")

    def test_forward_backward(self, rtol=1e-5):
        """Test that forward and backward transform give the right result"""

        r = np.random.randn(self.m.nz, self.m.ny, self.m.nx)
        r -= r.mean(axis=(-2,-1))[:,np.newaxis,np.newaxis]
        rh = self.m.fft(r)
        rf = self.m.ifft(rh)
        self.assertTrue(np.allclose(r, rf, rtol=rtol))

class BTModelFFTTester(ModelFFTMixin, unittest.TestCase):

    def setUp(self):
        self.m = pyqg.BTModel(L=2.*np.pi, beta=0., U=0., rd=0., filterfac=0)

class QGModelFFTTester(ModelFFTMixin, unittest.TestCase):

    def setUp(self):
        self.m = pyqg.QGModel(L=2.*np.pi, beta=0., U1=0., U2=0., filterfac=0.)


if __name__ == "__main__":
    test_parseval()
